// Copyright (c) 2018-2020 MobileCoin Inc.

//!  data structures not defined elsewhere.

use crate::mealy::{Input as MealyInput, Output as MealyOutput};
use aead::{AeadMut, NewAead};
use alloc::vec::Vec;
use core::marker::PhantomData;
use digest::{BlockInput, FixedOutput, Input, Reset};
use mc_attest_core::VerificationReport;
use mc_crypto_keys::Kex;
use mc_crypto_noise::{HandshakePattern, NoiseCipher, ProtocolName};

/// An input used to inject the relevant local data needed to transform Start
/// into an AuthPending for node-to-node authentication.
pub struct NodeInitiate<KexAlgo, Cipher, DigestType>
where
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    /// This is the local node's identity key
    pub(crate) local_identity: KexAlgo::Private,
    /// This is the local node's ias report.
    pub(crate) ias_report: VerificationReport,

    _kex: PhantomData<fn() -> KexAlgo>,
    _cipher: PhantomData<fn() -> Cipher>,
    _digest: PhantomData<fn() -> DigestType>,
}

impl<KexAlgo, Cipher, DigestType> NodeInitiate<KexAlgo, Cipher, DigestType>
where
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    /// Create a new input event to initiate a node-to-node channel.
    pub fn new(local_identity: KexAlgo::Private, ias_report: VerificationReport) -> Self {
        Self {
            local_identity,
            ias_report,
            _kex: PhantomData::default(),
            _cipher: PhantomData::default(),
            _digest: PhantomData::default(),
        }
    }
}

impl<KexAlgo, Cipher, DigestType> MealyInput for NodeInitiate<KexAlgo, Cipher, DigestType>
where
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
}

/// An input used to transform a Start into an AuthPending for client-to-node
/// authentication.
pub struct ClientInitiate<KexAlgo, Cipher, DigestType>
where
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    _kex: PhantomData<fn() -> KexAlgo>,
    _cipher: PhantomData<fn() -> Cipher>,
    _digest: PhantomData<fn() -> DigestType>,
}

impl<KexAlgo, Cipher, DigestType> Default for ClientInitiate<KexAlgo, Cipher, DigestType>
where
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    fn default() -> Self {
        Self {
            _kex: PhantomData::default(),
            _cipher: PhantomData::default(),
            _digest: PhantomData::default(),
        }
    }
}

impl<KexAlgo, Cipher, DigestType> MealyInput for ClientInitiate<KexAlgo, Cipher, DigestType>
where
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
}

/// An AuthRequest is an opaque blob of noise protocol handshake bytes,
/// generated by an initiator, and consumed by a responder.
pub struct AuthRequestOutput<Handshake, KexAlgo, Cipher, DigestType>
where
    Handshake: HandshakePattern,
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    /// The actual AuthRequest data
    pub(crate) data: Vec<u8>,

    /// Type consumption
    _protocol_name: ProtocolName<Handshake, KexAlgo, Cipher, DigestType>,
}

impl<Handshake, KexAlgo, Cipher, DigestType> From<Vec<u8>>
    for AuthRequestOutput<Handshake, KexAlgo, Cipher, DigestType>
where
    Handshake: HandshakePattern,
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    fn from(data: Vec<u8>) -> Self {
        Self {
            data,
            _protocol_name: ProtocolName::<Handshake, KexAlgo, Cipher, DigestType>::default(),
        }
    }
}

impl<Handshake, KexAlgo, Cipher, DigestType> Into<Vec<u8>>
    for AuthRequestOutput<Handshake, KexAlgo, Cipher, DigestType>
where
    Handshake: HandshakePattern,
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    fn into(self) -> Vec<u8> {
        self.data
    }
}

impl<Handshake, KexAlgo, Cipher, DigestType> AsRef<[u8]>
    for AuthRequestOutput<Handshake, KexAlgo, Cipher, DigestType>
where
    Handshake: HandshakePattern,
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    fn as_ref(&self) -> &[u8] {
        self.data.as_ref()
    }
}

/// An authentication request is output from an initiator
impl<Handshake, KexAlgo, Cipher, DigestType> MealyOutput
    for AuthRequestOutput<Handshake, KexAlgo, Cipher, DigestType>
where
    Handshake: HandshakePattern,
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
}

/// An input used to transform a Start into a Ready for a node responder.
///
/// It contains the AuthRequestOutput generated by an initiator, and the
/// supporting data a responder will need to complete the handshake.
pub struct AuthRequestInput<Handshake, KexAlgo, Cipher, DigestType>
where
    Handshake: HandshakePattern,
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    /// This is the local node's identity key
    pub(crate) local_identity: KexAlgo::Private,
    /// This is the local node's ias report.
    pub(crate) ias_report: VerificationReport,

    /// The auth request input, including payload, if any
    pub(crate) data: AuthRequestOutput<Handshake, KexAlgo, Cipher, DigestType>,
}

impl<Handshake, KexAlgo, Cipher, DigestType> MealyInput
    for AuthRequestInput<Handshake, KexAlgo, Cipher, DigestType>
where
    Handshake: HandshakePattern,
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
}

impl<Handshake, KexAlgo, Cipher, DigestType>
    AuthRequestInput<Handshake, KexAlgo, Cipher, DigestType>
where
    Handshake: HandshakePattern,
    KexAlgo: Kex,
    Cipher: AeadMut + NewAead + NoiseCipher + Sized,
    DigestType: BlockInput + Clone + Default + FixedOutput + Input + Reset,
{
    pub fn new(
        data: AuthRequestOutput<Handshake, KexAlgo, Cipher, DigestType>,
        local_identity: KexAlgo::Private,
        ias_report: VerificationReport,
    ) -> Self {
        Self {
            local_identity,
            ias_report,
            data,
        }
    }
}

/// An opaque blob output by a responder to complete a noise handshake.
pub struct AuthResponse(Vec<u8>);

impl From<Vec<u8>> for AuthResponse {
    fn from(src: Vec<u8>) -> Self {
        Self(src)
    }
}

impl AsRef<[u8]> for AuthResponse {
    fn as_ref(&self) -> &[u8] {
        self.0.as_ref()
    }
}

impl Into<Vec<u8>> for AuthResponse {
    fn into(self) -> Vec<u8> {
        self.0
    }
}

/// An authentication request is output from an initiator
impl MealyOutput for AuthResponse {}

/// An authentication request is input to a responder
impl MealyInput for AuthResponse {}

/// The output of an initiator when the responder has successfully
/// authenticated itself and the key exchange is complete.
pub type AuthSuccess = ();

/// Authentication success is output from an initiator
impl MealyOutput for AuthSuccess {}

/// A type similar to aead::Payload used to distinguish writer inputs from outputs.
pub struct Plaintext<'aad, 'msg> {
    pub aad: &'aad [u8],
    pub msg: &'msg [u8],
}

impl<'aad, 'msg> Plaintext<'aad, 'msg> {
    pub fn new(aad: &'aad [u8], msg: &'msg [u8]) -> Self {
        Self { aad, msg }
    }
}

/// Plaintext may be provided to an FST for encryption into a vector
impl MealyInput for Plaintext<'_, '_> {}

/// A type similar to aead::Payload used to distinguish reader inputs from outputs.
pub struct Ciphertext<'aad, 'msg> {
    pub aad: &'aad [u8],
    pub msg: &'msg [u8],
}

impl<'aad, 'msg> Ciphertext<'aad, 'msg> {
    pub fn new(aad: &'aad [u8], msg: &'msg [u8]) -> Self {
        Self { aad, msg }
    }
}

/// A ciphertext may be provided to a FST for decryption into a vector
impl MealyInput for Ciphertext<'_, '_> {}

/// Our outputs may be simple vectors for the proto-inside-grpc use case.
impl MealyOutput for Vec<u8> {}
